# IMPORTED FROM RTSS09

import os, sys, glob, shutil

import cpu2000
from cpu2000 import spec_dist, DefaultBenchmark
from config import *

SRC_SPEC_DIST = os.path.join(ROOT, 'spec/benchspec')


class qsort(DefaultBenchmark):
    name = "qsort"
    number = 1 
    lang = 'C'
    alternate_directory = "mibench/automotive/qsort"
    input_list = ['input_large.dat', 'input_small.dat'] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['input_small.dat']

    ref = test



class susan(DefaultBenchmark):
    name = "susan"
    number = 2 
    lang = 'C'
    alternate_directory = "mibench/automotive/susan"
    input_list = ['input_large.pgm', 'input_small.pgm'] 
    stdin = None
    stdout = None

    def test(self, isa, os):
        self.args = ['input_small.pgm', 'output_small.smoothing.pgm', '-s']

    ref = test



class basicmath(DefaultBenchmark):
    name = "basicmath"
    number = 3 
    lang = 'C'
    alternate_directory = "mibench/automotive/basicmath"
    input_list = [] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = []

    ref = test



class bitcount(DefaultBenchmark):
    name = "bitcount"
    number = 4 
    lang = 'C'
    alternate_directory = "mibench/automotive/bitcount"
    input_list = [] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['75000']

    ref = test

class lame(DefaultBenchmark):
    name = "lame"
    number = 5 
    lang = 'C'
    alternate_directory = "mibench/consumer/lame/lame3.70"
    input_list = ['../small.wav']
    stdin = None
    stdout = None
    extra_libs = '-L. -lmp3lame'

    def test(self, isa, os):
        self.args = ['small.wav', 'output_small.mp3']

    ref = test



class cjpeg(DefaultBenchmark):
    name = "cjpeg"
    number = 6 
    lang = 'C'
    alternate_directory = "mibench/consumer/jpeg/jpeg-6a"
    input_list = ['../input_small.ppm', '../input_small.jpg']
    stdin = None
    stdout = None

    def test(self, isa, os):
        self.args = ['-dct', 'int', '-progressive', '-opt', 
                '-outfile', 'output_small_encode.jpeg', 'input_small.ppm']

    ref = test


class djpeg(DefaultBenchmark):
    name = "djpeg"
    number = 1
    lang = 'C'
    alternate_directory = "jpeg-6b"
    input_list = [ "test.jpg" ]

    def test(self, isa, os):
        self.args = ["test.jpg", "test.bmp"]

    ref = test


class typeset(DefaultBenchmark):
    name = "typeset"
    number = 7 
    lang = 'C'
    alternate_directory = "mibench/consumer/typeset/lout-3.24"
    input_list = ['../lout-3.24/hyph', '../lout-3.24/font', 
        '../small.lout', '../lout-3.24/data', 
        '../lout-3.24/include', '../lout-3.24/maps'] 
    stdin = None
    stdout = 'output_small.ps'

    def test(self, isa, os):
        self.args = ['-I', 'include', '-D', 'data', 
                '-F', 'font', '-C', 'maps', 
                '-H', 'hyph', 'small.lout']

    ref = test



class CRC32(DefaultBenchmark):
    name = "CRC32"
    number = 8 
    lang = 'C'
    alternate_directory = "mibench/telecomm/CRC32"
    input_list = ['../adpcm/data/large.pcm'] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['large.pcm']

    ref = test

class fft(DefaultBenchmark):
    name = "fft"
    number = 9 
    lang = 'C'
    alternate_directory = "mibench/telecomm/FFT"
    input_list = [ 'fft' ] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['4', '4096']

    ref = test



class adpcm(DefaultBenchmark):
    name = "adpcm"
    number = 10 
    lang = 'C'
    alternate_directory = "mibench/telecomm/adpcm/src"
    input_list = [ '../data/small.pcm' ] 
    stdin = 'small.pcm'
    stdout = 'output_small.adpcm'

    def test(self, isa, os):
        self.args = []

    ref = test



class gsm(DefaultBenchmark):
    name = "gsm"
    number = 11 
    lang = 'C'
    alternate_directory = "mibench/telecomm/gsm/src"
    input_list = ['../data/small.au'] 
    stdin = None
    stdout = 'output_small.encode.gsm'

    def test(self, isa, os):
        self.args = ['-fps', '-c', 'small.au']

    ref = test



class sha(DefaultBenchmark):
    name = "sha"
    number = 12 
    lang = 'C'
    alternate_directory = "mibench/security/sha"
    input_list = ['input_large.asc', 'input_small.asc'] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['input_small.asc']

    ref = test



class rijndael(DefaultBenchmark):
    name = "rijndael"
    number = 13 
    lang = 'C'
    alternate_directory = "mibench/security/rijndael"
    input_list = ['input_large.asc', 'input_small.asc' ]
    stdin = None
    stdout = None

    def test(self, isa, os):
        self.args = ['input_small.asc', 'output_small.enc', 'e', '1234567890abcdeffedcba09876543211234567890abcdeffedcba0987654321']

    ref = test



class bf(DefaultBenchmark):
    name = "bf"
    number = 14 
    lang = 'C'
    alternate_directory = "mibench/security/blowfish"
    input_list = ['input_large.asc', 'input_small.asc' ]
    stdin = None
    stdout = None

    def test(self, isa, os):
        self.args = ['e', 'input_small.asc', 'output_small.enc', '1234567890abcdeffedcba0987654321']

    ref = test

class gs(DefaultBenchmark):
    name = "gs"
    number = 15 
    lang = 'C'
    alternate_directory = "mibench/office/ghostscript/src"
    input_list = ['../data/small.ps', '../../ghostscript'] 
    stdin = None
    stdout = None

    def test(self, isa, os):
        self.args = ['-Ighostscript', '-sDEVICE=ppm', 
            '-dNOPAUSE', '-q', '-sOutputFile=output_small.ppm', 
            '--', 'small.ps']

    ref = test



class rsynth(DefaultBenchmark):
    name = "rsynth"
    number = 16 
    lang = 'C'
    alternate_directory = "mibench/office/rsynth"
    input_list = ['smallinput.txt', 'largeinput.txt'] 
    stdin = 'smallinput.txt'
    stdout = None

    def test(self, isa, os):
        self.args = ['-a', '-q', '-o', 'small_output.au']

    ref = test



class ispell(DefaultBenchmark):
    name = "ispell"
    number = 17 
    lang = 'C'
    alternate_directory = "mibench/office/ispell"
    input_list = [ 'tests' ] 
    stdin = 'tests/small.txt'
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['-a', '-d', 'tests/americanmed+']

    ref = test



class stringsearch(DefaultBenchmark):
    name = "stringsearch"
    number = 18 
    lang = 'C'
    alternate_directory = "mibench/office/stringsearch"
    input_list = [] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = []

    ref = test



class patricia(DefaultBenchmark):
    name = "patricia"
    number = 19 
    lang = 'C'
    alternate_directory = "mibench/network/patricia"
    input_list = ['small.udp'] 
    stdin = None
    stdout = 'output_small.txt'

    def test(self, isa, os):
        self.args = ['small.udp']

    ref = test



class dijkstra(DefaultBenchmark):
    name = "dijkstra"
    number = 20 
    lang = 'C'
    alternate_directory = "mibench/network/dijkstra"
    input_list = ['input.dat'] 
    stdin = None
    stdout = 'output_small.dat'

    def test(self, isa, os):
        self.args = ['input.dat']

    ref = test

# class doom(DefaultBenchmark):
#     name = "linuxxdoom"
#     number = 2
#     lang = 'C'
#     alternate_directory = "linuxdoom-1.10"
#     input_list = [ "doom.wad" ]
# 
#     def test(self, isa, os):
#         self.args = []
# 
#     ref = test

class tester(DefaultBenchmark):
    name = "tester"
    number = 3
    lang = 'C'
    alternate_directory = "tester"
    input_list = []

    def test(self, isa, os):
        self.args = []

    ref = test

ALL_BENCHMARKS = (
        cpu2000.all + [ djpeg, tester ] +
        [basicmath, bitcount, susan, qsort ] +
        [ CRC32, typeset, cjpeg, lame ]  +
        [ bf, rijndael, sha, gsm, adpcm, fft, patricia, dijkstra ] +
        [ rsynth, ispell, stringsearch ]
        )

def On_Blacklist(bench):
    if ( bench.lang != "C" ):
        return True
    if ( bench.name in ("gcc", "perlbmk", "parser", "typeset") ):
        return True
    return False

def CCSetup():
    os.environ[ 'PATH' ] = ( os.path.join(GCC_BASE_DIR, "bin") + ":" +
                os.environ[ 'PATH' ])
    os.environ[ 'CC' ] = "%s-gcc" % SP
    os.environ[ 'LD' ] = "%s-gcc" % SP
    os.environ[ 'GGFN' ] = ' '.join([
        "-g", "-mcpu=ev56",
        "-fno-inline", "-fno-unroll-loops",
        "-fno-omit-frame-pointer", 
        "-fcse-follow-jumps", "-fcse-skip-blocks", 
        "-frerun-cse-after-loop", "-fgcse", 
        "-fcprop-registers",
        "-fexpensive-optimizations", "-fstrict-aliasing",
        "-fgcse-lm", "-fpeephole2",
        "-fsched-interblock", "-fsched-spec", "-fschedule-insns",
        "-fschedule-insns2", "-fregmove" ])

def Main(clean):
    # Verify that I'm in the directory that I expect.
    root = ROOT
    frontline = os.path.join(root, 'frontline')
    os.chdir(frontline)
    incdir = "/"
    os.chdir(incdir)
    os.chdir(SRC_SPEC_DIST)

    # Clean up an earlier build if any
    try:
        os.chdir(spec_dist)
        os.system("rm -rf data binaries")
    except:
        pass
    os.chdir(root)

    # Setup for C compiler
    CCSetup()

    done = set()

    def SpecialMake(bench_name, benchdir, input_list, extra_libs):
        # This make function will work for both SPEC benchmarks
        # and my own "other" benchmarks.


        os.chdir(benchdir)
        try:
            os.unlink('inc')
        except:
            pass
        #os.symlink(incdir, 'inc')

        rc = os.system("make clean")

        if ( clean ):
            return 0

        print ""
        print benchdir, bench_name
        rc = os.system("make %s" % bench_name)
        if ( rc != 0 ):
            print 'Make failed'
            print benchdir
            return rc

        print benchdir, bench_name

        bindir = os.path.join(spec_dist, "binaries",
                        "alpha", "linux")
        MkdirP(bindir)

        for (offset, binext) in [ (1024, "1"), (1024 + 65536, "2") ]:
            binname = bench_name + binext
            link = {
                    "TBB" : frontline,
                    "1" : offset,
                    "2" : binname,
                    "3" : ' '.join(glob.glob("*.o")),
                    "CRTI_DIR" : CRTI_DIR,
                    "CRTBEGINT_DIR" : CRTBEGINT_DIR,
                    "binext" : binext,
                    "SP" : SP,
                    "extra_libs" : extra_libs,
                    }

            for cmd in [ 
                    "%(SP)s-gcc -DSPACE_OFFSET=%(1)u "
                        "-c %(TBB)s/alphacrt1.S -I%(TBB)s -o crt1.O" % link,
                    "%(SP)s-gcc -O3 -c %(TBB)s/alphammapfixup.c "
                        "-o alphammapfixup.O" % link,
#                   "%(SP)s-ld -static -g -o %(2)s "
#                       "crt1.O %(CRTI_DIR)s/crti.o "
#                       "%(CRTBEGINT_DIR)s/crtbeginT.o "
#                       "%(3)s "
#                       "%(TBB)s/alphammapfixup.o "
#                       "-L%(CRTBEGINT_DIR)s -L%(CRTI_DIR)s -lc -lm -lgcc "
#                       "%(CRTBEGINT_DIR)s/crtend.o "
#                       "%(CRTI_DIR)s/crtn.o " % link,
                    "%(SP)s-gcc -static -g -o %(2)s "
                        "-specs=%(TBB)s/alphaspecs "
                        "%(3)s "
                        "-L%(CRTBEGINT_DIR)s -L%(CRTI_DIR)s "
                        "%(extra_libs)s -lc -lm "
                        "-lgcc -Wl,-T -Wl,%(TBB)s/linkscript" % link,
                    ]:

                print >>sys.stderr, ''
                print >>sys.stderr, cmd
                rc = os.system(cmd)
                if ( rc != 0 ):
                    print >>sys.stderr, 'Failed, rc %d' % rc
                    return rc

            Symlink(os.path.join(benchdir, binname), 
                    os.path.join(bindir, binname))
        
        Symlink(os.path.join(benchdir, binname), 
                os.path.join(bindir, bench_name))

        for input_set in [ "ref", "test", "train" ]:
            for (indirname, outdirname) in [ 
                        ("output", "expect"),
                        ("../all/input", "input"),
                        ("input", "input"),
                    ]:
                basedir = outdir = os.path.join(spec_dist, "data", 
                            bench_name, input_set)
                outdir = os.path.join(outdir, outdirname)
                indir = os.path.join(benchdir, "..", "data", 
                            input_set, indirname)
                MkdirP(outdir)

                if ( os.path.exists(indir) ):
                    for name in os.listdir(indir):
                        name2 = os.path.join(indir, name)
                        if ( os.path.isfile(name2) ):
                            Symlink(name2, os.path.join(outdir, name))


            outdir = os.path.join(basedir, "output")
            MkdirP(outdir)
            for name in input_list:
                src = os.path.join(benchdir, name)
                dest = os.path.join(basedir, "input", 
                            os.path.basename(name))
                if ( os.path.isdir(src) ):
                    try:
                        shutil.copytree(src, dest)
                    except:
                        print 'copytree failed'
                        print 'src', src
                        print 'dest', dest
                        raise
                else:
                    Symlink(src, dest)

            if ( bench_name == "equake" ):
                Symlink(os.path.join(basedir, "input", "inp.in"),
                    os.path.join(basedir, "input", "equake.in"))
                                
            os.chdir(os.path.join(basedir, "expect"))
            for name in glob.glob("*"):
                file(os.path.join(basedir, "output", name), "wt").close()
            os.chdir(benchdir)

        return 0


    for bench in ALL_BENCHMARKS:
        if (On_Blacklist(bench)):
            continue
        if ( bench.name in done ):
            continue

        done.add(bench.name)

        benchdir = getattr(bench, "alternate_directory", None)
        if ( benchdir == None ):
            os.chdir(SRC_SPEC_DIST)
            intdir = glob.glob("CINT2000/*.%s/src" % bench.name)
            fpdir = glob.glob("CFP2000/*.%s/src" % bench.name)
            if ( len(intdir) == 1 ):
                benchdir = intdir.pop()
            elif ( len(fpdir) == 1 ):
                benchdir = fpdir.pop()
            else:
                print 'bench %s not found!' % bench.name
                return 1

            benchdir = os.path.join(SRC_SPEC_DIST, benchdir)
        else:
            benchdir = os.path.join(root, benchdir)

        assert os.path.exists(benchdir), benchdir

        rc = SpecialMake(bench.name, benchdir, 
                    getattr(bench, "input_list", []),
                    getattr(bench, "extra_libs", ""))
        if ( rc != 0 ):
            return rc


def MkdirP(dlist):
    os.system("mkdir -p '%s'" % dlist)

def Symlink(src, dest):
    try:
        os.symlink(src, dest)
    except:
        print 'symlink failed'
        print 'src', src
        print 'dest', dest
        raise

if ( __name__ == "__main__" ):
    clean = ( "clean" in sys.argv )
    sys.exit(Main(clean))


