
import pickle, math, sys
import readcfg, ilp, ipet, stuff

def Main():
    size_list = [4096, 8192, 16384]
    limit_list = [8, 16]
    backtrack_list = [False, True]

    sys.setrecursionlimit(5000)

    # Gather phase 1 - experimental data
    tasks = pickle.load(file("s2tasks.dat", "rb"))
    name_set = set()
    collect = dict()
    key_list = []
    for key in tasks.iterkeys():
        (name, size, backtrack, limit) = key
        debug_dir = stuff.Make_Debug_Dir(name)
        identifier = "%s-%u-%s-%u" % (name, size, str(backtrack), limit)
        optlog_file = "%s/optlog-%s" % (debug_dir, identifier)

        try:
            ol = file(optlog_file, "rt")
        except:
            print 'missing', optlog_file
            continue

        ok = False
        main_memory = final = all_sp = "FAIL"
        for line in ol:
            line = line[ 5: ]
            if (line.startswith("main memory only iio wcet") ):
                main_memory = int(line.split()[ -1 ])
            elif (line.startswith("final wcet") ):
                final = int(line.split()[ -1 ])
            elif (line.startswith("all scratchpad iio wcet") ):
                all_sp = int(line.split()[ -1 ])
            elif (line.startswith("SCORE")):
                ok = True

        if not ok:
            print 'not ok ', optlog_file
            continue
        if name in ("tester", "sha", "bitcount"):
            print 'blcklst', optlog_file
            continue

        name_set.add(name)
        dynamic_ops_wcet = main_memory - all_sp
        perf = float(final - all_sp) / float(dynamic_ops_wcet)
        collect[ key ] = '%1.6f' % perf
        collect[ (name, 'wcets') ] = (all_sp, dynamic_ops_wcet)
        collect[ (key, 'main_memory') ] = main_memory
        collect[ (key, 'final') ] = final
        collect[ (key, 'all_sp') ] = all_sp
        key_list.append(key)

        print 'confirm', optlog_file

    name_list = list(name_set)
    name_list.sort(key=lambda x: x.lower())
    key_list.sort()

    key_sub = dict()
    for name in name_list:
        best_value = None
        best_key = None
        for size in size_list:
            key = (name, size, True, 16)
            value = collect.get((key, 'final'), None)
            if (value == None):
                pass
            elif ((best_value == None)
            or (value <= best_value)):
                best_value = value 
                best_key = key
            elif (value > best_value):
                key_sub[ key ] = best_key


    # Gather phase 2 - information about programs
    try:
        prog_info = pickle.load(file("progs.dat", "rb"))
    except:
        prog_info = Gather_Program_Data(name_list, key_list)
        pickle.dump(prog_info, file("progs.dat", "wb"))

    collect.update(prog_info)

    # Readout for Table \ref{stats} and \ref{inittable}
    name_list_2 = []
    stats = file("z/stats.tex", "wt")
    inittable = file("z/inittable.tex", "wt")
    mmm = file("z/mmm.tex", "wt")
    bfits = file("z/b.tex", "wt")
    both = Combined([stats, inittable, mmm, bfits])
    stats.write(r'\begin{tabular}{|l|rrr|rr|}')
    inittable.write(r'\begin{tabular}{|l|r|rr|r|r|}') 
    mmm.write(r'\begin{tabular}{|l|rrrrr|}')  # XXX
    bfits.write(r'\begin{tabular}{|l|rrrr|}')  # XXX
    both.write('\n')
    both.write(r'\hline')

    both.write('\n')
    both.write(r'\ts Benchmark')
    mmm.write(r'&\multicolumn{5}{c|}{\ts $s_p$ Is Greater Than} \\')
    mmm.write('\n')
    bfits.write(r'&\multicolumn{4}{c|}{\ts $B(p)$ Is Greater Than} \\')
    bfits.write('\n')
    stats.write(r'&\multicolumn{3}{c|}{\ts Proportion of WCEP}&')
    stats.write(r'\multicolumn{2}{c|}{\ts WCET} \\')
    stats.write('\n')

    for i in xrange(1):
        inittable.write(r'& \ts Prop. & \ts Max & \ts Max')
        inittable.write(r'& \ts E & \multirow{2}{*}{\ts \rratio} ')
    inittable.write(r'\\')
    inittable.write('\n')

    for i in xrange(1):
        inittable.write(r'& \ts DS & \ts OPEN & \ts size & ')
    inittable.write(r'\\')

    mmm.write(r'\ts & 5\% & \ts 25\% & \ts 50\% & \ts 75\% & \ts 95\% \\')
    bfits.write(r'\ts & \ts 25\% & \ts 50\% & \ts 75\% & \ts 95\% \\')
    PERC = [5, 25, 50, 75, 95]

    stats.write(r'\ts & \ts R & \ts S & \ts D & \ts Code & \ts DE \\')
    both.write('\n')
    both.write(r'\hline')
    both.write('\n')
    total_non_memory = []
    all_ds = []
    all_ratio = []

    for name in name_list:
        key = (name, 16384, True, 16)
        key = key_sub.get(key, key)
        result2 = collect.get(key, None)
        if result2 == None:
            continue

        name_list_2.append(name)
        both.write(r'\ts %14s' % name.lower())
        (prop_non_memory, prop_static, prop_dynamic, total) = collect[ name ]
        for v in (prop_non_memory, prop_static, prop_dynamic):
            v *= 100.0
            Two_SF(stats, v)
            stats.write(r"\% ")

        total_non_memory.append(prop_non_memory)

        (all_sp, dynamic_ops_wcet) = collect[ (name, 'wcets') ]
        for v in (all_sp, dynamic_ops_wcet):
            v /= 1e6
            Two_SF(stats, v)
            stats.write(r" ")

        prop_external = collect.get((key, 'prop_external'), None)
        assert (prop_external != None), key
        v = 1.0 - prop_external
        v *= 100.0
        Two_SF(inittable, v)        # DS
        all_ds.append(v)
        inittable.write(r"\% ")

        inittable.write(r"&\ts %4d " % collect[ (key, 'max_open') ])
        inittable.write(r"&\ts %5d " % collect[ (key, 'max_open_size') ])

        main_memory = collect[ (key, 'main_memory') ]
        final = collect[ (key, 'final') ]
        all_sp = collect[ (key, 'all_sp') ]

        external = float(final - all_sp)
        v = external / 1e6
        Two_SF(inittable, v)       
        inittable.write(r" ")

        ratio = (external / (main_memory - all_sp))
        all_ratio.append(ratio)
        assert 0 <= ratio <= 1.0
        v = ratio 
        Two_SF(inittable, v)       
        inittable.write(r" ")


        percentiles = collect[ (key, 'percentiles') ]
        for pc in PERC:
            bytes = percentiles[ pc - 1 ]
            kilo = 1024
            meg = kilo * kilo
            if ( bytes >= meg ):
                mmm.write(r"&\ts %4uMb " % (bytes / meg))
            elif ( bytes >= kilo ):
                mmm.write(r"&\ts %4ukb " % (bytes / kilo))
            else:
                mmm.write(r"&\ts %4ub  " % (bytes))
        
        benefits = collect[ (key, 'benefits') ]
        for pc in PERC[ 1: ]:
            v = benefits[ pc - 1 ]
            basis = math.floor(math.log10(abs(v)))
            v = float(v) / math.pow(10, basis)
            bfits.write(r'&\ts %5.1f$\times 10^{\mbox{\tiny %3.0f}}$' % (v, basis))
        
        both.write(r'\\')
        both.write('\n')
        both.flush()

    both.write(r'\hline')
    both.write('\n')
    both.write(r'\end{tabular}')
    both.write('\n')
    both.close()

    name_list = name_list_2
    print 'mean total non memory = %1.2f%%' % (100.0 *
                sum(total_non_memory) / len(total_non_memory))
    print 'mean DS = %1.2f%%' % (sum(all_ds) / len(all_ds))
    print 'mean ratio = %1.3f' % (
                sum(all_ratio) / len(all_ratio))

    # Readout phase

    size_list = [16384, 8192, 4096]
    limit_list = [16]
    backtrack_list = [True]

    csv = file("z/initgraph.csv", "wt")
    for size in reversed(size_list):
        csv.write(',%u' % size)
    csv.write("\n")

    for name in name_list:
        csv.write(name.lower())
        for size in reversed(size_list):
            key = (name, size, True, 16)
            key = key_sub.get(key, key)
            csv.write(',%s' % collect.get(key, ''))
        csv.write("\n")
    csv.close()


    grid = dict()
    x = 0
    for size in size_list:
        for limit in limit_list:
            for backtrack in backtrack_list:
                grid[ (x, -3) ] = str(size)
                grid[ (-1, -3) ] = 'size'
                grid[ (x, -2) ] = str(limit)
                grid[ (-1, -2) ] = 'limit'
                grid[ (x, -1) ] = str(backtrack)
                grid[ (-1, -1) ] = 'bt'
                x += 1

    for element in [ "non_memory", "static", "dynamic" ]:
        grid[ (x, -1) ] = element
        x += 1
           
    y = 0
    for name in name_list:
        grid[ (-1, y) ] = name
        x = 0
        for size in size_list:
            for limit in limit_list:
                for backtrack in backtrack_list:
                    key = (name, size, backtrack, limit)
                    key = key_sub.get(key, key)
                    grid[ (x, y) ] = collect.get(key, '')
                    x += 1
        for (i, element) in enumerate([ "non_memory", "static", "dynamic" ]):
            grid[ (x, y) ] = '%1.1f%%' % (100.0 * collect[ name ][ i ])
            x += 1
        y += 1


    min_x = -1
    min_y = -3
    max_x = x
    max_y = y

    width = [ 0 for x in xrange(min_x, max_x, 1) ]

    for y in xrange(min_y, max_y, 1):
        for x in xrange(min_x, max_x, 1):
            width[ x ] = max(len(grid.get((x, y), '')), width[ x ])

    for y in xrange(min_y, max_y, 1):
        for x in xrange(min_x, max_x, 1):
            print grid.get((x, y), '').rjust(width[ x ]),
        print ''

class My_Base_Creator_2(stuff.My_Base_Creator):
    def __init__(self, *args):
        stuff.My_Base_Creator.__init__(self, *args)
        self.usage_events = 0
        self.create_events = 0

        
def Gather_Program_Data(name_list, key_list):
    prog_info = dict()
    for name in name_list:
        debug_dir = stuff.Make_Debug_Dir(name)
        cfg_file = "%s/cfg2" % (debug_dir)
        cfg_entrance = readcfg.Read_CFG(cfg_file,
            edge_class=stuff.My_CFG_Edge,
            base_creator_class=My_Base_Creator_2)
        loop_header_list = readcfg.Identify_Loops(cfg_entrance) # Required for IPET
        model = ilp.ILP_Solver("%s/mdl.%s.infoprog" % (debug_dir, name))
        (vertex_list, edge_list) = readcfg.Find_V_E(cfg_entrance)

        alloc_obj = stuff.Allocation_Object(-1, 1, 1)
        edge_lookup = dict()

        for edge in edge_list:
            edge.alloc_obj = alloc_obj
            edge_lookup[ edge.id ] = edge

        bc_list = readcfg.Find_Base_Creators(cfg_entrance)
        bc_heap_list = []
        bc_lookup = dict()
        for bc in bc_list:
            if (bc.Important()):
                bc_heap_list.append(bc)
                bc_lookup[ bc.id ] = bc

        stuff.Recalc_Costs(bc_heap_list, edge_list)
        best_wcet = ipet.Simple_IPET(cfg_entrance, model)

        count_dynamic = 0
        count_static = 0
        count_non_memory = 0

        for edge in edge_list:
            heap_ops = set()
            for (bc, pc_list) in edge.user:
                if bc.Important():
                    for pc in pc_list:
                        heap_ops.add(pc)

            for pc in edge.path:
                if (pc in edge.mem_ops):
                    if (pc in heap_ops):
                        count_dynamic += edge.frequency
                    else:
                        count_static += edge.frequency
                else:
                    assert not pc in heap_ops
                    count_non_memory += edge.frequency

        total = float(count_dynamic + count_static + count_non_memory)
        prog_info[ name ] = (count_non_memory / total, 
                    count_static / total, count_dynamic / total, total)
        print name, count_non_memory, count_static, count_dynamic

        # Now read in the output file
        for key in key_list:
            (name1, size, backtrack, limit) = key
            if (name != name1):
                continue
            if (not backtrack):
                continue

            identifier = "%s-%u-%s-%u" % (name, size, str(backtrack), limit)
            oc_out_file = "%s/oc-%s" % (debug_dir, identifier)
            alloc_obj_list = []
            alloc_obj = None
            open_count = open_size = 0
            max_open = max_open_size = 0

            for line in file(oc_out_file, "rt"):
                fields = line.split()
                k = value = None
                if (len(fields) >= 2):
                    k = fields[ 0 ]
                    try:
                        value = int(fields[ 1 ], 16)
                    except:
                        value = fields[ 1 ]

                if (k == "ao"):
                    alloc_obj = stuff.Allocation_Object(value, size, limit)
                    max_open = max(open_count, max_open)
                    max_open_size = max(open_size, max_open_size)
                    alloc_obj_list.append(alloc_obj)
                    open_count = 0
                    open_size = 0
                elif (k == "edge"):
                    edge_lookup[ value ].alloc_obj = alloc_obj
                    alloc_obj.edge_list.append(edge)
                elif (k == "alloc"):
                    if (fields[ -1 ] != "empty"):
                        open_count += 1
                        bc = bc_lookup[ int(fields[ -1 ], 16) ]
                        rc = alloc_obj.Allocate_At(bc, value)
                        assert rc
                elif (k == "s"):
                    open_size = value


            for bc in bc_heap_list:
                bc.usage_events = 0
                bc.create_events = 0

            stuff.Recalc_Costs(bc_heap_list, edge_list)
            best_wcet = ipet.Simple_IPET(cfg_entrance, model)

            count_dynamic = 0
            count_external = 0
            total_oc = 0
            access_size = dict()

            for edge in edge_list:
                heap_ops = set()
                ext_ops = set()
                max_size = 0
                total_oc += edge.open_close_cost * edge.frequency

                for bc in edge.creator:
                    bc.create_events += edge.frequency

                for (bc, pc_list) in edge.user:
                    if bc.Important():
                        for pc in pc_list:
                            heap_ops.add(pc)
                            if not edge.alloc_obj.BC_Is_Open(bc):
                                ext_ops.add(pc)
                            max_size = max(max_size, bc.size)
                            bc.usage_events += edge.frequency

                for pc in edge.path:
                    if (pc in ext_ops):
                        count_external += edge.frequency
                    if (pc in heap_ops):
                        count_dynamic += edge.frequency
                        access_size[ max_size ] = access_size.get(max_size, 0) + edge.frequency

            prop_external = float(count_external) / float(count_dynamic)
            prog_info[ (key, 'prop_external') ] = prop_external
            prog_info[ (key, 'max_open') ] = max_open
            prog_info[ (key, 'max_open_size') ] = max_open_size
            prog_info[ (key, 'total_oc') ] = total_oc
            print '   ', size, backtrack, limit, '  ', 
            print max_open, max_open_size, '  %1.3f' % prop_external

            benefits = []
            for bc in bc_heap_list:
                if bc.create_events != 0:
                    benefits.append((
                        ((bc.usage_events * stuff.MAIN_MEMORY_ACCESS_COST) - (
                            bc.create_events * bc.Tx_Cost()) - bc.usage_events) / bc.create_events, 
                        bc.usage_events))
            benefits.sort()

            prog_info[ (key, 'benefits') ] = Compile_Percentiles(benefits)

            size_to_count = access_size.items()
            size_to_count.sort()
            prog_info[ (key, 'percentiles') ] = Compile_Percentiles(size_to_count)

    return prog_info

def Compile_Percentiles(size_to_count):
    total_count = 0
    for (size, count) in size_to_count:
        total_count += count
    
    stop_at = [ ((i + 1) * total_count) / 100 for i in xrange(100) ]
    stop_at.reverse()
    
    total_count = 0
    percentiles = []
    for (size, count) in size_to_count:
        total_count += count
        while ((len(stop_at) != 0)
        and (total_count >= stop_at[ -1 ])):
            percentiles.append(size)
            stop_at.pop()

    print '    ', percentiles
    assert len(percentiles) == 100
    return percentiles

def Two_SF(stats, v):
    if ( v <= 1.0 ):
        stats.write(r'& \ts %4.2f' % v)
    elif ( v <= 10.0 ):
        stats.write(r'& \ts %4.1f' % v)
    else:
        d = 1.0
        if ( v >= 1000.0 ):
            d = 100.0
        elif ( v >= 100.0 ):
            d = 10.0
        stats.write(r'& \ts %4.0f' % (math.floor((v + (d / 2)) / d) * d))

class Combined:
    def __init__(self, out):
        self.out = out
        print self.out

    def write(self, text):
        for out in self.out:
            out.write(text)

    def flush(self):
        for out in self.out:
            out.flush()

    def close(self):
        for out in self.out:
            out.close()

if __name__ == "__main__":
    Main()

